/* Y o u r   D e s c r i p t i o n                       */
/*                            AppBuilder Photon Code Lib */
/*                                         Version 1.13A */

/* Standard headers */
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

/* Toolkit headers */
#include <Ph.h>
#include <Pt.h>
#include <Ap.h>

/* Local headers */
#include "chatrooms.h"
#include "abimport.h"
#include "proto.h"


void setUserData( const PtWidget_t *wgt, const void *ptr )
	{
	PtArg_t	args[ 1 ];

	PtSetArg( &args[ 0 ], Pt_ARG_USER_DATA, &ptr, sizeof( void * ) );
	PtSetResources( wgt, 1, args );

	}	// setUserData


void * getUserData( const PtWidget_t *wgt )
	{
		PtArg_t		args[ 1 ];
static	void		*ptr;

	PtSetArg( &args[ 0 ], Pt_ARG_USER_DATA, &ptr, sizeof( void * ) );
	PtGetResources( wgt, 1, args );

	return ptr;
	}	// getUserData


typedef struct
	{
	PtWidget_t	*win;
	PtWidget_t	*balloon;
	char		*text;
	int			position;
	}	BALLOON;


int
widget_balloon( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )
	{
	BALLOON		*ptr;

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	/*--- Nothing to do ? ---*/
	if ( cbinfo->event->type != Ph_EV_BOUNDARY )
		{
		ptr = *(BALLOON**)getUserData( widget );
		if ( ptr->balloon )
			{
			PtDestroyWidget( ptr->balloon );
			ptr->balloon = NULL;
			}
		return( Pt_CONTINUE );
		}

	/*--- Enter or Leave this grid cell ? ---*/
	switch ( cbinfo->event->subtype )
		{
		
		case Ph_EV_PTR_ENTER :
			ptr = *(BALLOON**)getUserData( widget );
			ptr->balloon = PtInflateBalloon( ABW_room, widget, ptr->position, ptr->text, 
				"helv12", Pg_BALLOONCOLOR, Pg_BLACK );
			break;

		case Ph_EV_PTR_LEAVE :
			ptr = *(BALLOON**)getUserData( widget );
			if ( ptr->balloon )
				PtDestroyWidget( ptr->balloon );
			break;
		
		}	// switch

	return( Pt_CONTINUE );

	}	// widget_balloon


int install_balloons( PtWidget_t *win )
	{
#ifdef NEVER
	PtArg_t		args[ 1 ];
	PtWidget_t	*curr;
	char		*ptr_help_text;
	BALLOON		*ptr;
static	PtRawCallback_t	cb_raw[] = 
	{
	{ Ph_EV_BOUNDARY | Ph_EV_BUT_PRESS | Ph_EV_KEY, widget_balloon, NULL }
	};

	curr = win;	
	while ( (curr = PtWidgetFamily( win, curr )) != NULL )
		{

		PtSetArg( &args[ 0 ], Pt_ARG_HELP_TOPIC, &ptr_help_text, 0 );
		PtGetResources( curr, 1, args );
		if ( ptr_help_text == NULL )
			continue;

		PtSetArg( &args[ 0 ], Pt_CB_RAW, cb_raw, sizeof( cb_raw ) / sizeof( cb_raw[ 0 ] ) );
		PtSetResources( curr, 1, args );

		setUserData( curr, ptr = malloc( sizeof( BALLOON ) ) );
		ptr->win = win;
		ptr->text = strdup( &ptr_help_text[ 1 ] );

		switch ( ptr_help_text[ 0 ] )
			{
			case 'L' : ptr->position = Pt_BALLOON_LEFT;		break;
			case 'R' : ptr->position = Pt_BALLOON_RIGHT;	break;
			case 'T' : ptr->position = Pt_BALLOON_TOP;		break;
			case 'B' : ptr->position = Pt_BALLOON_BOTTOM;	break;
			case 'I' : 
			default  : ptr->position = Pt_BALLOON_INPLACE;	break;
			}
		ptr->balloon = NULL;
		
		}	// while
#endif

	return( Pt_CONTINUE );

	}	// install_ballons


int desinstall_balloons( PtWidget_t *win )
	{
#ifdef NEVER
	PtArg_t		args[ 1 ];
	PtWidget_t	*curr;
	char		*ptr_help_text;
	BALLOON		*ptr;

	curr = win;	
	while ( (curr = PtWidgetFamily( win, curr )) != NULL )
		{

		PtSetArg( &args[ 0 ], Pt_ARG_HELP_TOPIC, &ptr_help_text, 0 );
		PtGetResources( curr, 1, args );
		if ( ptr_help_text == NULL )
			continue;

		ptr = getUserData( curr );
		ptr = *(BALLOON**)ptr;
		free( ptr->text );
		free( ptr );
		
		}	// while
#endif

	return( Pt_CONTINUE );

	}	// desinstall_ballons

