/* Y o u r   D e s c r i p t i o n                       */
/*                            AppBuilder Photon Code Lib */
/*                                         Version 1.13A */

/* Standard headers */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <process.h>
#include <sys/sched.h>
#include <sys/qnx_glob.h>
#include <sys/name.h>
#include <unistd.h>
#include <sys/kernel.h>
#include <signal.h>
#include <sys/proxy.h>
#include <sys/psinfo.h>
#include <assert.h>
#include <sys/proc_msg.h>
#include <pwd.h>

/* Toolkit headers */
#include <Ph.h>
#include <Pt.h>
#include <Ap.h>

/* Local headers */
#include "chatrooms.h"
#include "abimport.h"
#include "proto.h"

#define __MAIN__
#include "chatrooms_vars.h"

/* Application Options string */
const char ApOptions[] =
	AB_OPTIONS ""; /* Add your options in the "" */


extern int send_msg( void *data, pid_t pid, void *msg, ushort_t size );


void update_chairs( USER *p0 )
{
	USER		*p;
	int			nb_users;
	int			chair;
	PhPoint_t	new_pos;
	int			y;
	PtArg_t		args[1];
	PhDim_t		*old_dim, new_dim;

	/* Nb of users into the current room and hide some chairs */
	for ( nb_users = 0, p = p0, chair = 0; chair < MAXNB_CHAIRS; chair++, p++ )
		if ( strcmp( p, "" ) )
			nb_users++;

	/* Resize and reposition the chairs */
	for ( y = 91, p = p0, chair = 0; chair < MAXNB_CHAIRS; chair++, p++ ) 
		{
		new_pos.x = 6;
		if ( strcmp( p, "" ) ) 
			{
			new_pos.y = y;
			y += (75 + 2);
			} 
		else
			{
			new_pos.y = 10000;
			}
		PtSetArg( &args[0], Pt_ARG_POS, &new_pos, 0 );
		PtSetResources( room_chairs[ chair ], 1, args );
		}	// for (n)

	/*--- Resize the window ---*/
	PtSetArg( &args[0], Pt_ARG_DIM, &old_dim, 0 );
	PtGetResources( ABW_room, 1, args );
	memmove( &new_dim, old_dim, sizeof( PhDim_t ) );
	new_dim.h = 112 + (75+2) * nb_users;
	PtSetArg( &args[0], Pt_ARG_DIM, &new_dim, 0 );
	PtSetResources( ABW_room, 1, args );

}	// update_chairs


void active_text( const short chair, char *user_name, const int active )
{
	PtArg_t		args[ 2 ];

	PtSetArg( &args[0], Pt_ARG_TEXT_FLAGS, 
		(chair == curr_chair) ? Pt_TRUE : Pt_FALSE, Pt_EDITABLE );
	PtSetArg( &args[ 0 ], Pt_ARG_CURSOR_TYPE,
		(chair == curr_chair) ? Ph_CURSOR_INSERT : Ph_CURSOR_NOINPUT, 0 );
	PtSetResources( room_texts[ chair ], 2, args );

	PtSetArg( &args[0], Pt_ARG_TEXT_STRING, user_name, 0 );
	PtSetResources( room_labels[ chair ], 1, args );

}	// active_text


void show_users( void )
{
	USER			*q;
	int				room, chair;
	PtArg_t			args[1];

	/* Update the chairs */
	if ( ABW_room && (curr_room != NO_ROOM) )
		update_chairs( all_users[ curr_room ] );

	/* Refresh the list of users into each room */
	for ( q = all_users, room = 0; room < MAXNB_ROOMS; room++ ) {
		for ( chair = 0; chair < MAXNB_CHAIRS; chair++ ) {
			PtSetArg (&args[0], Pt_ARG_TEXT_STRING, q, 0);
			PtSetResources( chatrooms_chair[ room ][ chair ], 1, args );
			if ( ABW_room && (room == curr_room) )
				active_text( chair, q, strcmp( q, "" ) );
			q++;
		}
	}

}	// show_users


int users_change( void *data, pid_t pid, void *msg, ushort_t size )
{
	REQ_GETWHO			req;
	ANSWER_GETWHO		answer;
	USER				*p, *q;
	int					room, chair;
	PtArg_t				args[1];
	int					update;
	int					status;

	/* eliminate 'unreferenced' warnings */
	data= data, pid= pid, msg= msg, size= size;

	/* Send the request to the server */	
	req.hdr.type = RQ_GETWHO;
	req.userno   = userno;
	status=	Send( server_vid, &req, &answer, 
		sizeof( REQ_GETWHO ), sizeof( ANSWER_GETWHO ) );
	assert( status == 0 );

	/* Update the chairs */
	if ( ABW_room && (curr_room != NO_ROOM) )
		update_chairs( answer.users[ curr_room ] );

	/* Refresh the list of users into each room */
	for ( update = FALSE, p = answer.users, q = all_users, room = 0; 
		  room < MAXNB_ROOMS; 
		  room++ ) {
		for ( chair = 0; chair < MAXNB_CHAIRS; chair++ ) {
			if ( strcmp( p, q ) ) {
				PtSetArg (&args[0], Pt_ARG_TEXT_STRING, p, 0);
				PtSetResources( chatrooms_chair[ room ][ chair ], 1, args );
				if ( ABW_room && (room == curr_room) ) {
					update = TRUE;
					active_text( chair, p, strcmp( p, "" ) );
				}	// if
				strcpy( q, p );
			}	// if
			p++;
			q++;
		}	// for (chair)
	}	// for (room)
	
	return( Pt_CONTINUE );

}	// users_change


int show_msg( void )
{
	MSG			*q;
	int			chair;
	PtArg_t		args[1];

	/* Refresh the list of msg into each chair */
	for ( q = all_chairs_msg, chair = 0; chair < MAXNB_CHAIRS; chair++, q++ ) {
		PtSetArg (&args[0], Pt_ARG_TEXT_STRING, q, 0);
		PtSetResources( room_texts[ chair ], 1, args );
	}	// for (chair)
	
	return( Pt_CONTINUE );

}	// show_msg


int msg_change( void *data, pid_t pid, void *msg, ushort_t size )
{
	REQ_GET_ROOM_MSG	req;
	ANSWER_GET_ROOM_MSG	answer;
	MSG					*p, *q;
	int					chair;
	PtArg_t				args[1];
	int					status;

	/* eliminate 'unreferenced' warnings */
	data= data, pid= pid, msg= msg, size= size;

	/* Send the request to the server */	
	req.hdr.type = RQ_GET_ROOM_MSG;
	req.userno   = userno;
	status = Send( server_vid, &req, &answer, 
		sizeof( REQ_GET_ROOM_MSG ), sizeof( ANSWER_GET_ROOM_MSG ) );
	assert( status == 0 );

	/* Refresh the list of msg into each chair */
	for ( p = answer.room_msg, q = all_chairs_msg, chair = 0; chair < MAXNB_CHAIRS; chair++ ) {
		if ( (chair != curr_chair) && strcmp( p, q ) ) {
			PtSetArg (&args[0], Pt_ARG_TEXT_STRING, p, 0);
			PtSetResources( room_texts[ chair ], 1, args );
			strcpy( q, p );
			}
		p++;
		q++;
	}	// for (chair)
	
	return( Pt_CONTINUE );

}	// msg_change


int invitation( char *from_who, nid_t from_node, int which_room )
{
	REQ_ENTER_ROOM		req;
	ANSWER_ENTER_ROOM	answer;
	int					status;
	struct sigevent		sigevnt;

	/* Message sent to the server */
	req.hdr.type = RQ_ENTER_ROOM;
	req.userno   = userno;
	req.room     = which_room;
	status = Send( server_vid, &req, &answer, 
		sizeof( REQ_ENTER_ROOM ), sizeof( ANSWER_ENTER_ROOM ) );
	assert( status == 0 );
	if ( answer.ack == ACK_FULL ) {
		setString( ABW_chatrooms_infos, "Room full !" );
		return( Pt_CONTINUE );
	}
	curr_room  = which_room;
	curr_chair = answer.chair;
	memmove( all_users, answer.users, sizeof( all_users ) );
	memmove( all_chairs_msg, answer.room_msg, sizeof( answer.room_msg ) );

	/* Refresh the list of users */
	show_users();

	/* Function signaled after a timeout to send the msg to the server */
	proxy_send_msg = qnx_proxy_attach( 0, 0, 0, -1 );
	sigevnt.sigev_signo = -proxy_send_msg;
	timer_send_msg = timer_create( CLOCK_REALTIME, &sigevnt );
	send_msg_id = PtAppAddInput( NULL, proxy_send_msg, send_msg, NULL );

	/* --- Open the room window --- */
	ApCreateModule( ABM_room, NULL, NULL );
	setString( ABW_room_infos, "You were invited by [%s] from node %d",
		from_who, from_node );
	setString( ABW_chatrooms_infos, "" );

	return( Pt_CONTINUE );

}	// invitation


int exit_app( int signal, void *data )
{
	REQ_DECONNECT		req;
	ANSWER_DECONNECT	answer;
	int					status;

	if ( signal==SIGTERM || signal==SIGINT ) {
		req.hdr.type = RQ_DECONNECT;
		req.userno   = userno;
		status = Send( server_vid, &req, &answer, 
			sizeof( REQ_DECONNECT ), sizeof( ANSWER_DECONNECT ) );
		exit( 0 );
	}

	return( Pt_CONTINUE );

}	// exit_app


static int set_sid_name( int sid, char *name )
	{
	union 
		{
		struct _proc_session s;
		struct _proc_session_reply r;
		}	msg;

	msg.s.type = _PROC_SESSION;
	msg.s.subtype = _PROC_SUB_ACTION1;
	msg.s.sid = sid;
	strcpy( msg.s.name, name );

	if ( Send( PROC_PID, &msg.s, &msg.r, sizeof(msg.s), sizeof(msg.r) ) == -1 )
		return -1;
	if ( msg.r.status )
		{
		errno = msg.r.status;
		return -1;
		}
	return 0;
	}	// set_sid_name

int
chatrooms_init( int argc, char *argv[] )
{
	char 			path[PATH_MAX + 1], exename[PATH_MAX + 1];
	char 			_node[_MAX_NODE + 1], _dir[_MAX_DIR + 1], _fname[_MAX_FNAME + 1],
					_ext[_MAX_EXT + 1];
	REQ_CONNECT		req;
	ANSWER_CONNECT	answer;
	int				status;
	struct _psinfo	psinfos;
	nid_t			server_nid;
	sigset_t		set;
	struct passwd	*pw;

	/* Already running ? */
#if SIMULATION==0
	if ( (qnx_name_attach( 0, CLIENT_NAME )) == -1 )
		exit( -1 );
#endif

	setsid();
	memset( whoami, 0, MAXLEN_USER+1 );
	if ( argc == 7 )
		strncpy( whoami, argv[ 5 ], MAXLEN_USER );
	else
		strncpy( whoami, getlogin(), MAXLEN_USER );
	set_sid_name( getsid( getpid() ), whoami );

	/*--- Which home directory ? ---*/
	memset( ini_fname, 0, sizeof( ini_fname ) );
	if ( (pw = getpwnam( whoami )) != NULL )
		{
		char	p[ PATH_MAX+1 ];
		int		l;
		strcpy( p, pw->pw_dir );
		if ( ((l=strlen(p)) >= 1) && (p[l-1]=='/') )
			p[l-1] = 0;
		sprintf( ini_fname, "%s/.photon/%s", p, CFG_FNAME );
		}

	/* Create the proxies to be signaled on users/msg events */
	if ( ((proxy_users = qnx_proxy_attach( 0, 0, 0, -1)) == -1) ||
		 ((proxy_msg = qnx_proxy_attach( 0, 0, 0, -1)) == -1) ) {
		fprintf( stderr, "No proxy available!\n" );
		exit( -1 );
	}

	/* Try to either locate or lauch the server */
	if ( (server_vid = qnx_name_locate( 0, SERVER_NAME, 
		max( sizeof( UNION_REQ_MSG ), sizeof( UNION_ANSWER_MSG ) ), 0 )) == -1 ) {
		if (_cmdname(path) == NULL)
			return( Pt_CONTINUE );
		_splitpath( path, _node, _dir, _fname, _ext );
		sprintf(exename, "%s%s", _dir, "chatrooms_server" );
		qnx_spawn_options.flags |= _SPAWN_NOZOMBIE;
		server_vid = spawnl( P_NOWAIT, exename, exename, NULL );
	}

	/* Server on which node ? */
	qnx_psinfo( PROC_PID, server_vid, &psinfos, 0, 0 );
	if ( psinfos.flags & _PPF_VID )
		server_nid = psinfos.un.vproc.remote_nid;
	else
		server_nid = getnid();

	/* Global init */
	curr_room = NO_ROOM;
	userno = -1;
	memset( all_users, 0, sizeof( all_users ) );
	ABW_room  = NULL;
	ABW_invit = NULL;

	/* Function called on exit application */
	sigemptyset( &set );
	sigaddset( &set, SIGINT );
	sigaddset( &set, SIGQUIT );
	sigaddset( &set, SIGTERM );
	PtAppAddSignalProc( NULL, &set, exit_app, NULL );

	/* Function called when there is change in users */
	PtAppAddInput( NULL, proxy_users, users_change, NULL );
	PtAppAddInput( NULL, proxy_msg, msg_change, NULL );

	/* This is an invitation ? */
	if ( argc == 7 ) {
		extern struct _Ph_ctrl	*_Ph_;
		char	tmp[ 80];
		if ( strcmp( argv[1], "GlopGlop" ) )
			exit( -1 );
		PhDetach( _Ph_ );
		sprintf( tmp, "//%d/dev/photon", getnid() );
		if ( PhAttach( tmp, NULL ) == NULL )
			exit( 0 );
		PhEventArm();
	}

	/* Sent a connect message to the server */
	memset( &req, 0, sizeof( REQ_CONNECT ) );
	req.hdr.type = RQ_CONNECT;
	strcpy( req.user, whoami );
	req.node = getnid();
	req.pid  = getpid();
	req.vproxy_users = qnx_proxy_rem_attach( server_nid, proxy_users );
	req.vproxy_msg   = qnx_proxy_rem_attach( server_nid, proxy_msg );
	status = Send( server_vid, &req, &answer, 
		sizeof( REQ_CONNECT ), sizeof( ANSWER_CONNECT ) );
	assert( status == 0 );
	userno = answer.userno;
	memmove( all_users, &answer.users, sizeof( all_users ) );

	/* This is an invitation ? */
	if ( argc == 7 ) {
		ApCreateModule( ABM_chatrooms, NULL, NULL );
		PtUnrealizeWidget( ABW_chatrooms );
		chatrooms_was_hidden = TRUE;
		invitation( argv[ 2 ], atoi( argv[ 3 ] ), atoi( argv[ 4 ] ) );
		return( Pt_CONTINUE );
	} else {

		/* --- Open the 'chatrooms' window --- */
		ApCreateModule( ABM_chatrooms, NULL, NULL );
		chatrooms_was_hidden = FALSE;

	}

	return( Pt_CONTINUE );

}	// chatrooms_init


int
chatrooms_setup( PtWidget_t *link_instance, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )

{

	/* eliminate 'unreferenced' warnings */
	link_instance = link_instance, apinfo = apinfo, cbinfo = cbinfo;

#define DEF( room, chair ) \
	chatrooms_chair[ room ][ chair ] = ABW_room##room##_chair##chair;
	DEF( 0, 0 ); DEF( 0, 1 ); DEF( 0, 2 ); DEF( 0, 3 ); DEF( 0, 4 );
	DEF( 1, 0 ); DEF( 1, 1 ); DEF( 1, 2 ); DEF( 1, 3 ); DEF( 1, 4 );
	DEF( 2, 0 ); DEF( 2, 1 ); DEF( 2, 2 ); DEF( 2, 3 ); DEF( 2, 4 );
	DEF( 3, 0 ); DEF( 3, 1 ); DEF( 3, 2 ); DEF( 3, 3 ); DEF( 3, 4 );
#undef DEF

	show_users();

	return( Pt_CONTINUE );

}	// chatrooms_setup


int
chatroom_activate( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )
{
	PtArg_t 			args[1];
	char				*ptr;
	REQ_ENTER_ROOM		req;
	ANSWER_ENTER_ROOM	answer;
	struct sigevent		sigevnt;
	int					status;

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	/* Which room has been selected ? */
	PtSetArg( &args[0], Pt_ARG_USER_DATA, &ptr, sizeof( void *) );
	PtGetResources( widget, 1, args );
	curr_room = atoi( ptr );

	/* Message sent to the server */
	req.hdr.type = RQ_ENTER_ROOM;
	req.userno   = userno;
	req.room     = curr_room;
	status = Send( server_vid, &req, &answer, 
		sizeof( REQ_ENTER_ROOM ), sizeof( ANSWER_ENTER_ROOM ) );
	assert( status == 0 );
	if ( answer.ack == ACK_FULL ) {
		setString( ABW_chatrooms_infos, "Room full !" );
		return( Pt_CONTINUE );
	}
	curr_chair = answer.chair;
	memmove( all_users, answer.users, sizeof( all_users ) );
	memmove( all_chairs_msg, answer.room_msg, sizeof( answer.room_msg ) );

	/* Refresh the list of users */
	show_users();

	/* Function signaled after a timeout to send the msg to the server */
	proxy_send_msg = qnx_proxy_attach( 0, 0, 0, -1 );
	sigevnt.sigev_signo = -proxy_send_msg;
	timer_send_msg = timer_create( CLOCK_REALTIME, &sigevnt );
	send_msg_id = PtAppAddInput( NULL, proxy_send_msg, send_msg, NULL );

	/* --- Open the 'room' window --- */
	ApCreateModule( ABM_room, NULL, NULL );
	setString( ABW_chatrooms_infos, "" );

	return( Pt_CONTINUE );

}	// chatroom_activate


int
chatroom_boundary( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )
{
	PtArg_t 	args[3];
	int			state;

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	/* Nothing to do ? */
	if ( ABW_room != NULL )		// Already into a room
		if ( cbinfo->event->type != Ph_EV_BOUNDARY )
			return( Pt_CONTINUE );

	if ( cbinfo->event->subtype == Ph_EV_PTR_ENTER )
		state = -1;
	else if ( cbinfo->event->subtype == Ph_EV_PTR_LEAVE )
		state = 0;
	else
		return( Pt_CONTINUE );
	
	PtSetArg( &args[0], Pt_ARG_FLAGS, (state) ? Pt_TRUE : Pt_FALSE, Pt_SET );
	PtSetArg( &args[1], Pt_ARG_FLAGS, (state) ? Pt_TRUE : Pt_FALSE, Pt_ETCH_HIGHLIGHT );
	PtSetResources( widget, 2, args );

	return( Pt_CONTINUE );

}	// chatroom_boundary


int
chatrooms_btn_exit( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )
{

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	/* Exit the application */
	kill( getpid(), SIGTERM );

	return( Pt_CONTINUE );

}	// chatrooms_btn_exit


int
chatrooms_destroyed( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )

	{

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	exit_app( SIGINT, NULL );

	return( Pt_CONTINUE );

	}	// chatrooms_destroyed
