#define DEBUG			1
#define SIMULATION		0
#define	LOG				1

// Version number of software
#define VERSION_NB		0x006301	// version 0.99.A
#define VERSION_DATE	0x000505	// yymmdd

#define	SERVER_NAME		"/qnx/chatrooms_server"
#define	CLIENT_NAME		"qnx/chatrooms"

#define MAXNB_ROOMS		4
#define	MAXNB_CHAIRS	5
#define	MAXNB_USERS		(MAXNB_ROOMS * MAXNB_CHAIRS)

#define	MAXLEN_USER		16
typedef	char USER[ MAXLEN_USER+1 ];

#define	MAXLEN_MSG		240
typedef char MSG[ MAXLEN_MSG+1 ];

#define SIG_USERS		SIGUSR1
#define SIG_MSG			SIGUSR2

#include <sys/sys_msg.h>

/* Requests which can be sent to the chatrooms_server */
enum {
	RQ_CONNECT = 0x1001,	// Initial connection to the server
	RQ_DECONNECT,			// Deconnexion from the server
	RQ_GETINFOS,			// Get some informations
	RQ_GETWHO,				// Get who is in each chatroom
	RQ_ENTER_ROOM,			// Enter in a given chatroom (0-3)
	RQ_EXIT_ROOM,			// Exit from a given chatroom (0-3)
	RQ_TYPE_MSG,			// Type a message when into a room
	RQ_GET_ROOM_MSG,		// Get all msg from the current room
};

/* Ack/Nack */
enum {
	ACK_OK,
	ACK_FULL,			// Room is full
};

/* In a room ? In a chair ? */
#define NO_ROOM		-1
#define	NO_CHAIR	-1

#define FALSE 0
#define TRUE  -1


/* ------------------------- */
/* Structures for RQ_CONNECT */
/* ------------------------- */
typedef struct {
	struct _sysmsg_hdr	hdr;
	USER	user;
	nid_t	node;
	pid_t	pid;
	pid_t	vproxy_users, vproxy_msg;
}	REQ_CONNECT;
typedef struct {
	struct _sysmsg_hdr_reply	hdr;
	short	ack;
	short	userno;		// User Identifier
	int		nb_users;
	USER	users[ MAXNB_ROOMS ][ MAXNB_CHAIRS ];
}	ANSWER_CONNECT;


/* --------------------------- */
/* Structures for RQ_DECONNECT */
/* --------------------------- */
typedef struct {
	struct _sysmsg_hdr	hdr;
	short	userno;
}	REQ_DECONNECT;
typedef struct {
	struct _sysmsg_hdr_reply	hdr;
	short	ack;
}	ANSWER_DECONNECT;


/* -------------------------- */
/* Structures for RQ_GETINFOS */
/* -------------------------- */
typedef struct {
	struct _sysmsg_hdr	hdr;
	short	userno;
}	REQ_GETINFOS;
typedef struct {
	struct _sysmsg_hdr_reply	hdr;
	short	ack;
	nid_t	server_node;
	int		nb_users;
}	ANSWER_GETINFOS;


/* ------------------------ */
/* Structures for RQ_GETWHO */
/* ------------------------ */
typedef struct {
	struct _sysmsg_hdr	hdr;
	short	userno;
}	REQ_GETWHO;
typedef struct {
	struct _sysmsg_hdr_reply	hdr;
	short	ack;
	int		nb_users;
	USER	users[ MAXNB_ROOMS ][ MAXNB_CHAIRS ];
}	ANSWER_GETWHO;


/* ---------------------------- */
/* Structures for RQ_ENTER_ROOM */
/* ---------------------------- */
typedef struct {
	struct _sysmsg_hdr	hdr;
	short	userno;
	short	room;						// Room number (0..3)
}	REQ_ENTER_ROOM;
typedef struct {
	struct _sysmsg_hdr_reply	hdr;
	short	ack;
	short	chair;						// In which chair the user is
	int		nb_users;
	USER	users[ MAXNB_ROOMS ][ MAXNB_CHAIRS ];
	MSG		room_msg[ MAXNB_CHAIRS ];	// The messages
}	ANSWER_ENTER_ROOM;


/* --------------------------- */
/* Structures for RQ_EXIT_ROOM */
/* --------------------------- */
typedef struct {
	struct _sysmsg_hdr	hdr;
	short	userno;
}	REQ_EXIT_ROOM;
typedef struct {
	struct _sysmsg_hdr_reply	hdr;
	short	ack;
	int		nb_users;
	USER	users[ MAXNB_ROOMS ][ MAXNB_CHAIRS ];
}	ANSWER_EXIT_ROOM;


/* -------------------------- */
/* Structures for RQ_TYPE_MSG */
/* -------------------------- */
typedef struct {
	struct _sysmsg_hdr	hdr;
	short	userno;
	MSG		msg;				// The message
}	REQ_TYPE_MSG;
typedef struct {
	struct _sysmsg_hdr_reply	hdr;
	short	ack;
}	ANSWER_TYPE_MSG;


/* ------------------------------ */
/* Structures for RQ_GET_ROOM_MSG */
/* ------------------------------ */
typedef struct {
	struct _sysmsg_hdr	hdr;
	short	userno;
}	REQ_GET_ROOM_MSG;
typedef struct {
	struct _sysmsg_hdr_reply	hdr;
	MSG		room_msg[ MAXNB_CHAIRS ];	// The messages
}	ANSWER_GET_ROOM_MSG;


/* -------------------------------------------------- */
/* Used only to compute the maximum size of a message */
/* -------------------------------------------------- */

typedef union {
	REQ_CONNECT			connect;
	REQ_DECONNECT		deconnect;
	REQ_GETINFOS		getinfos;
	REQ_GETWHO			getwho;
	REQ_ENTER_ROOM		enter_room;
	REQ_EXIT_ROOM		exit_room;
	REQ_TYPE_MSG		type_msg;
	REQ_GET_ROOM_MSG	room_msg;
}	UNION_REQ_MSG;

typedef union {
	ANSWER_CONNECT		connect;
	ANSWER_DECONNECT	deconnect;
	ANSWER_GETINFOS		getinfos;
	ANSWER_GETWHO		getwho;
	ANSWER_ENTER_ROOM	enter_room;
	ANSWER_EXIT_ROOM	exit_room;
	ANSWER_TYPE_MSG		type_msg;
	ANSWER_GET_ROOM_MSG	room_msg;
}	UNION_ANSWER_MSG;


/* ---------------------- */
/* Configuration filename */
/* ---------------------- */

#define	CFG_FNAME			"chatrooms.ini"
#define	CFG_SECTION_USERS	"users"
