#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/name.h>
#include <sys/kernel.h>
#include <sys/vc.h>
#include <signal.h>
#include <assert.h>
#include <unistd.h>
#include <sys/psinfo.h>
#include <sys/proxy.h>
#include <time.h>
#include <errno.h>
#include <stdarg.h>

#include "chatrooms.h"

static	char	request[ sizeof( UNION_REQ_MSG ) ];
static	char	answer[ sizeof( UNION_ANSWER_MSG ) ];
static	pid_t	pid_req;

static	USER	all_users[ MAXNB_ROOMS ][ MAXNB_CHAIRS ];
static	MSG		all_rooms_msg[ MAXNB_ROOMS ][ MAXNB_CHAIRS ];

static	int		nb_users;
typedef struct 
	{
	short	cnx;
	USER	user;
	short	room, chair;
	pid_t	pid_client;
	pid_t	vpid_client;
	pid_t	vproxy_users, vproxy_msg;
	}	USER_INFO;
static	USER_INFO	users_infos[ MAXNB_USERS ];

#define	TIMEOUT_NOACTIVITY	600
static	pid_t	proxy_noactivity;	// Proxy signaled after a timeout, if no user connected
static timer_t	timer_noactivity;


#if DEBUG==1
void print_state( int sig_number )
	{
	int			n;
	USER_INFO	*u;
	int			room, chair;
	MSG			*m;

	printf( "*** nb_users = %d\n", nb_users );

	for ( u = users_infos, n = 0; n < MAXNB_USERS; n++, u++ )
		if ( u->cnx )
			printf( "user %-2d: room=%d chair=%d [%s]\n", n, u->room, u->chair, u->user );

	for ( m = all_rooms_msg, room = 0; room < MAXNB_ROOMS; room++ )
		for ( chair = 0; chair < MAXNB_CHAIRS; chair++ ) 
			{
			if ( strcmp( m, "" ) )
				printf( "msg: room=%d chair=%d [%s]\n", room, chair, m );
			m++;
			}

	}	// print_state	
#endif


#if LOGTOTMP==1
static void slog( const char *fmt, ... )
	{
	va_list			ap;
	char			msg[ 300 ];
	FILE			*fp;
	time_t			now;
	struct tm		*tm;

	/*--- Develop text to log ---*/
	va_start( ap, fmt );
	vsprintf( msg, fmt, ap );
	va_end( ap );

	/*--- Update log file ---*/
	time( &now );
	tm = localtime( &now );
	if ( (fp = fopen( "/tmp/log.txt", "a" )) != NULL )
		{
		fprintf( fp, "%02d:%02d:%02d [chat] %s\n",
			tm->tm_hour, tm->tm_min, tm->tm_sec,
			msg );
		fclose( fp );
		}

	}	// slog
#endif


static void change_in_users( const short not_this_user )
	{
	USER_INFO	*user_info;
	int			userno;

	for ( user_info = users_infos, userno = 0; 
		  userno < MAXNB_USERS; 
		  userno++, user_info++ )
		if ( user_info->cnx && (userno != not_this_user) )
			Trigger( user_info->vproxy_users );

	}	// change_in_users


static void new_msg( const short room, const short not_this_chair )
{
	USER_INFO	*user_info;
	int			n;

	for ( user_info = users_infos, n = 0; 
		  n < MAXNB_USERS; 
		  n++, user_info++ )
		if ( user_info->cnx && (user_info->room == room) && (user_info->chair != not_this_chair) )
			Trigger( user_info->vproxy_msg );

}	// new_msg


static void rq_connect( void )
{
	REQ_CONNECT		*prequest = (REQ_CONNECT *)request;
	ANSWER_CONNECT	*panswer = (ANSWER_CONNECT *)answer;
	short			userno;
	USER_INFO		*user_info;

	/* All rooms full */
	if ( nb_users == (MAXNB_ROOMS * MAXNB_CHAIRS) ) {
		panswer->ack = ACK_FULL;
		panswer->userno = -1;
		Reply( pid_req, panswer, sizeof( ANSWER_CONNECT ) );
		return;
		}

	/* Attrib a user id to this new connected */
	panswer->ack = ACK_OK;
	for ( user_info = users_infos, userno = 0; userno < MAXNB_USERS; userno++, user_info++ )
		if ( !user_info->cnx )
			break;
	// assert
	panswer->userno = userno;
	panswer->nb_users = nb_users;
	memmove( panswer->users, all_users, sizeof( all_users ) );
	Reply( pid_req, panswer, sizeof( ANSWER_CONNECT ) );

	/* Create virtual circuit with the client */
	user_info->vproxy_users = prequest->vproxy_users;
	user_info->vproxy_msg   = prequest->vproxy_msg;
	user_info->vpid_client  = qnx_vc_attach( prequest->node, prequest->pid,
		max( sizeof( UNION_REQ_MSG ), sizeof( UNION_ANSWER_MSG ) ), 0 );
	user_info->pid_client   = pid_req;

	/* Update table */
	strcpy( user_info->user, prequest->user );
	user_info->cnx    = TRUE;
	user_info->room   = NO_ROOM;
	user_info->chair  = NO_CHAIR;
	nb_users++;
#if LOGTOTMP==1
	slog( "[%s] connect", prequest->user );
#endif	

}	// rq_connect


#if SIMULATION==1
static void creat_user( int room, int chair, char *username, char *msg )
{
	short			userno;
	USER_INFO		*user_info;

	/* Attrib a user id to this new connected */
	for ( user_info = users_infos, userno = 0; userno < MAXNB_USERS; userno++, user_info++ )
		if ( !user_info->cnx )
			break;

	/* Update table */
	user_info->vproxy_users = 0;
	user_info->vproxy_msg   = 0;
	user_info->vpid_client  = 0;
	user_info->pid_client   = 0;
	strcpy( user_info->user, username );
	strcpy( all_users[ room ][ chair ], username );
	user_info->cnx    = TRUE;
	user_info->room   = room;
	user_info->chair  = chair;
	nb_users++;

	/* Update the list of all msg */
	strcpy( all_rooms_msg[ room ][ chair ], msg );

}	// creat_user
#endif


static void rq_deconnect( void )
{
	REQ_DECONNECT		*prequest = (REQ_DECONNECT *)request;
	ANSWER_DECONNECT	*panswer  = (ANSWER_DECONNECT *)answer;
	USER_INFO			*user_info;
	short				room, chair;
	struct itimerspec	timerSpec;

	/* Reset the msg from this user into this room,chair, if any */
	user_info = &users_infos[ prequest->userno ];
	room  = user_info->room;
	chair = user_info->chair;
#if LOGTOTMP==1
	slog( "[%s] deconnect", user_info->user );
#endif	
	if ( (room != NO_ROOM) && (chair != NO_CHAIR) ) {
		memset( &all_rooms_msg[ room ][ chair ], 0, sizeof( MSG ) );
		memset( all_users[ room ][ chair ], 0, sizeof( USER ) );
	}
	qnx_vc_detach( user_info->vpid_client );
	user_info->cnx = FALSE;
	memset( user_info->user, 0, sizeof( USER ) );
	nb_users--;

	/* Reply to the client */	
	panswer->ack = ACK_OK;
	Reply( pid_req, panswer, sizeof( ANSWER_DECONNECT ) );

	/* Signal all other clients that there is a change in the list of users into rooms */
	change_in_users( prequest->userno );
	
	/* Signal all the clients into this room */
	if ( (room != NO_ROOM) && (chair != NO_CHAIR) )
		new_msg( room, chair );

	/* Arm the timer to end the server after a while */
	if ( !nb_users ) {
		timerSpec.it_value.tv_sec     = TIMEOUT_NOACTIVITY;
		timerSpec.it_value.tv_nsec    =
		timerSpec.it_interval.tv_nsec =
		timerSpec.it_interval.tv_sec  = 0;
		timer_settime( timer_noactivity, 0, &timerSpec, NULL );
	}	// if

}	// rq_deconnect


static void rq_getinfos( void )
{
	ANSWER_GETINFOS	*panswer = (ANSWER_GETINFOS *)answer;

	/* Answer */	
	panswer->ack = ACK_OK;
	panswer->nb_users = nb_users;
	panswer->server_node = getnid();
	Reply( pid_req, panswer, sizeof( ANSWER_GETINFOS ) );
	
}	// rq_getinfos


static void rq_getwho( void )
{
	ANSWER_GETWHO	*panswer = (ANSWER_GETWHO *)answer;

	/* Answer */	
	panswer->ack = ACK_OK;
	panswer->nb_users = nb_users;
	memmove( panswer->users, all_users, sizeof( all_users ) );
	Reply( pid_req, panswer, sizeof( ANSWER_GETWHO ) );
	
}	// rq_getwho


static void rq_enter_room( void )
{
	REQ_ENTER_ROOM		*prequest = (REQ_ENTER_ROOM *)request;
	ANSWER_ENTER_ROOM	*panswer  = (ANSWER_ENTER_ROOM *)answer;
	USER_INFO			*user_info;
	short				room;
	int					chair;
	USER				*u;

	/* Room full ? */
	room = prequest->room;
	for ( u = all_users[ room ][ 0 ], chair = 0; chair < MAXNB_CHAIRS; chair++, u++ )
		if ( !strcmp( u, "" ) )
			break;
	if ( chair == MAXNB_CHAIRS ) {	
		panswer->ack = ACK_FULL;
		Reply( pid_req, panswer, sizeof( ANSWER_ENTER_ROOM ) );
		return;
		}

	/* Insert this user into this room */
	user_info = &users_infos[ prequest->userno ];
	user_info->room  = room;
	user_info->chair = chair;
	memset( u, 0, sizeof( USER ) );
	strcpy( u, user_info->user );
#if LOGTOTMP==1
	slog( "[%s] enter into room #%d", user_info->user, room+1 );
#endif	

	/* Reply to the client */	
	panswer->ack      = ACK_OK;
	panswer->chair    = chair;
	panswer->nb_users = nb_users;
	memmove( panswer->users, all_users, sizeof( all_users ) );
	memmove( panswer->room_msg, all_rooms_msg[ room ], sizeof( panswer->room_msg ) );
	Reply( pid_req, panswer, sizeof( ANSWER_ENTER_ROOM ) );

	/* Signal all other clients that there is a change in the list of users into rooms */
	change_in_users( prequest->userno );
	
}	// rq_enter_room


static void rq_exit_room( void )
{
	REQ_EXIT_ROOM		*prequest = (REQ_EXIT_ROOM *)request;
	ANSWER_EXIT_ROOM	*panswer = (ANSWER_EXIT_ROOM *)answer;
	USER_INFO			*user_info;
	short				room, chair;

	/* User not in a room ?? */
	user_info = &users_infos[ prequest->userno ];
	assert( (user_info->room >= 0) && (user_info->room < MAXNB_ROOMS) );

	/* Reset the msg from this user into this room */
	room  = user_info->room;
	chair = user_info->chair;
#if LOGTOTMP==1
	slog( "[%s] exit from the room #%d", user_info->user, room+1 );
#endif	

	memset( &all_rooms_msg[ room ][ chair ], 0, sizeof( MSG ) );

	/* Remove this user from this room */
	memset( all_users[ room ][ chair ], 0, sizeof( USER ) );
	user_info->room  = NO_ROOM;
	user_info->chair = NO_CHAIR;

	/* Reply to the client */	
	panswer->ack = ACK_OK;
	panswer->nb_users = nb_users;
	memmove( panswer->users, all_users, sizeof( all_users ) );
	Reply( pid_req, panswer, sizeof( ANSWER_EXIT_ROOM ) );

	/* Signal all other clients that there is a change in the list of users into rooms */
	change_in_users( prequest->userno );
	
	/* Signal all the clients into this room */
	new_msg( room, chair );

}	// rq_exit_room


static void rq_type_msg( void )
{
	REQ_TYPE_MSG		*prequest = (REQ_TYPE_MSG *)request;
	ANSWER_TYPE_MSG		*panswer = (ANSWER_TYPE_MSG *)answer;
	USER_INFO			*user_info;

	/* User not in a room ?? not int a chair ?? */
	user_info = &users_infos[ prequest->userno ];
	assert( (user_info->room >= 0) && (user_info->room < MAXNB_ROOMS) );
	assert( (user_info->chair >= 0) && (user_info->chair < MAXNB_CHAIRS) );

	/* Update the list of all msg */
	strcpy( all_rooms_msg[ user_info->room ][ user_info->chair ], prequest->msg );
	
	/* Reply to the client */	
	panswer->ack = ACK_OK;
	Reply( pid_req, panswer, sizeof( ANSWER_TYPE_MSG ) );

	/* Signal all the clients into this room */
	new_msg( user_info->room, user_info->chair );

}	// rq_type_msg


static void rq_get_room_msg( void )
{
	REQ_GET_ROOM_MSG	*prequest = (REQ_GET_ROOM_MSG *)request;
	ANSWER_GET_ROOM_MSG	*panswer  = (ANSWER_GET_ROOM_MSG *)answer;
	USER_INFO			*user_info;

	/* User not in a room ?? */
	user_info = &users_infos[ prequest->userno ];
	assert( (user_info->room >= 0) && (user_info->room < MAXNB_ROOMS) );

	/* Reply to the client */	
	memmove( panswer->room_msg, all_rooms_msg[ user_info->room ], sizeof( panswer->room_msg ) );
	Reply( pid_req, panswer, sizeof( ANSWER_GET_ROOM_MSG ) );

}	// rq_get_room_msg


static void sysmsg( const short subtype )
{
	struct _sysmsg_hdr_reply			*rdeath = (struct _sysmsg_hdr_reply *) answer;
	typedef struct {
		struct _sysmsg_hdr				hdr;
		struct _sysmsg_version_reply	data;
	}	VERSION_MSG;
	VERSION_MSG 						*rvers  = (VERSION_MSG *) answer;
	USER_INFO							*user_info;
	int									userno;
	int									found;
	short								room, chair;
	struct itimerspec					timerSpec;
	static char *months[ 12 ] = {
		"Jan", "Feb", "Mar", "Apr", "May", "Jun",
		"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
	};

	switch ( subtype ) {

		case _SYSMSG_SUBTYPE_DEATH :

			for ( found = FALSE, user_info = users_infos, userno = 0; 
				  userno < MAXNB_USERS; 
				  userno++, user_info++ )
				if ( user_info->cnx && (user_info->pid_client == pid_req) ) {
					found = TRUE;
					break;
				}

			if ( found ) {

				room  = user_info->room;
				chair = user_info->chair;
				if ( (room != NO_ROOM) && (chair != NO_CHAIR) ) {
					memset( &all_rooms_msg[ room ][ chair ], 0, sizeof( MSG ) );
					memset( all_users[ room ][ chair ], 0, sizeof( USER ) );
				}
#if LOGTOTMP==1
				slog( "[%s] deconnect by death of process", user_info->user );
#endif	
				user_info->cnx = FALSE;
				memset( user_info->user, 0, sizeof( USER ) );
				nb_users--;

				/* Signal all other clients that there is a change in the list of users into rooms */
				change_in_users( userno );

				/* Signal all the clients into this room */
				if ( (room != NO_ROOM) && (chair != NO_CHAIR) )
					new_msg( room, chair );

				/* Arm the timer to end the server after a while */
				if ( !nb_users ) {
					timerSpec.it_value.tv_sec     = TIMEOUT_NOACTIVITY;
					timerSpec.it_value.tv_nsec    =
					timerSpec.it_interval.tv_nsec =
					timerSpec.it_interval.tv_sec  = 0;
					timer_settime( timer_noactivity, 0, &timerSpec, NULL );
				}	// if

			}	// if

			rdeath->status = 0;
			Reply( pid_req, rdeath, sizeof( struct _sysmsg_hdr_reply ) );
			break;	// case _SYSMSG_SUBTYPE_DEATH

		case _SYSMSG_SUBTYPE_VERSION:

			rvers->hdr.type = EOK;
			strcpy( rvers->data.name, "Chatrooms" );
			rvers->data.version = ((VERSION_NB >> 16) * 100) +
				((VERSION_NB >> 8) & 0xFF);
			rvers->data.letter  = (VERSION_NB & 0xFF) + 'a';
			sprintf( rvers->data.date, "%s %02d %04d",
				months[ ((VERSION_DATE >> 8) & 0xFF) - 1 ], 
				(VERSION_DATE & 0xFF),
				(VERSION_DATE >> 16) + 1998 );
			rvers->data.more    = 0;

			Reply( pid_req, rvers, sizeof( VERSION_MSG ) );
			break;	// case _SYSMSG_SUBTYPE_VERSION
		
	}	// switch ( subtype )

}	// sysmsg


int main( int argc, char *argv[] )
{
	unsigned			copies;
	USER_INFO			*user_info;
	int					n;
	struct sigevent		sigevnt;
	struct itimerspec	timerSpec;
	struct _sysmsg_hdr	*msg_hdr = (struct _sysmsg_hdr *)request;
	int					finish;

	setsid();

	/* Register a name in order client will be able to locate the server */
	if ( qnx_name_locate( 0, SERVER_NAME, 0, NULL ) != -1 )
		return 0;
	if ( qnx_name_attach( 0, SERVER_NAME ) == -1 )
		return 0;
	if ( qnx_name_locate( 0, SERVER_NAME, 0, &copies ) == -1 )
		return 0;
	if ( copies > 1 )
		return 0;

	/* Global init */
	memset( all_users, 0, sizeof( all_users ) );
	memset( all_rooms_msg, 0, sizeof( all_rooms_msg ) );
	nb_users = 0;
	memset( users_infos, 0, sizeof( users_infos ) );
	for ( user_info = users_infos, n = 0; n < MAXNB_USERS; n++, user_info++ ) {
		user_info->cnx   = FALSE;
		user_info->room  = NO_ROOM;
		user_info->chair = NO_CHAIR;
	}

#if SIMULATION==1
	creat_user( 0, 0, "Oliver", "Hello everybody !" );
	creat_user( 0, 1, "Alain", "Bonjour !" );
	creat_user( 0, 2, "sascha", "Gutten tag !" );
	creat_user( 0, 3, "toto", "Hi guys !" );
	creat_user( 1, 0, "dtdodge", "QNX is great !\nThe best realtime OS on the market !" );
	creat_user( 1, 1, "gbell", "The only question is: why there are other OS !" );
	creat_user( 3, 0, "alone", "Nobody wants to talk with me !!!" );
#endif

	/* Function signaled after a timeout if no user connected */
	proxy_noactivity = qnx_proxy_attach( 0, 0, 0, -1 );
	sigevnt.sigev_signo = -proxy_noactivity;
	timer_noactivity = timer_create( CLOCK_REALTIME, &sigevnt );

	/* Set as Server */
	qnx_pflags( ~0, _PPF_SERVER | _PPF_INFORM, 0, 0 );

#if DEBUG==1
	signal( SIGUSR1, print_state );
#endif

	/* Handle requests */
	for ( finish = FALSE; !finish || (nb_users > 0); ) {


		if ( (pid_req = Receive( 0, request, sizeof( request ) )) == -1 )
			continue;

		if ( pid_req == proxy_noactivity ) {

			assert( (nb_users == 0 ) );
			finish = TRUE;

		} else {
		
			if ( msg_hdr->type == _SYSMSG ) {

				sysmsg( msg_hdr->subtype ); 
				
			} else {

				/* Reset timer */
				timerSpec.it_value.tv_sec     = 
				timerSpec.it_value.tv_nsec    = 
				timerSpec.it_interval.tv_sec  = 
				timerSpec.it_interval.tv_nsec = 0;
				timer_settime( timer_noactivity, 0, &timerSpec, NULL );

				switch ( msg_hdr->type ) {
					case RQ_CONNECT			: rq_connect(); break;
					case RQ_DECONNECT		: rq_deconnect(); break;
					case RQ_GETINFOS		: rq_getinfos(); break;
					case RQ_GETWHO			: rq_getwho(); break;
					case RQ_ENTER_ROOM		: rq_enter_room(); break;
					case RQ_EXIT_ROOM		: rq_exit_room(); break;
					case RQ_TYPE_MSG		: rq_type_msg(); break;
					case RQ_GET_ROOM_MSG	: rq_get_room_msg(); break;
				}	// switch
			}
		}
	
	}	// for (!finish)

	/* Done */
	qnx_pflags( 0, _PPF_SERVER | _PPF_INFORM, 0, 0 );
	return 0;

}	// main
