/* Y o u r   D e s c r i p t i o n                       */
/*                            AppBuilder Photon Code Lib */
/*                                         Version 1.13A */

/* Standard headers */
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <process.h>
#include <sys/sched.h>
#include <sys/qnx_glob.h>
#include <sys/name.h>
#include <unistd.h>
#include <sys/kernel.h>
#include <signal.h>
#include <sys/proxy.h>
#include <assert.h>
#include <sys/osinfo.h>
#include <sys/sidinfo.h>
#include <sys/vc.h>
#include <sys/types.h>

/* Toolkit headers */
#include <Ph.h>
#include <Pt.h>
#include <Ap.h>

/* Local headers */
#include "chatrooms.h"
#include "abimport.h"
#include "proto.h"


static	USER			who;

static	struct _osinfo	odata;
static	char			*nid_stat = NULL;


#include "chatrooms_vars.h"


static void blockWindow( const PtWidget_t *wgt, const int state )
{
	PtArg_t	args[ 2 ];

	if ( wgt ) {
		PtSetArg( &args[ 0 ], Pt_ARG_FLAGS, (state) ? Pt_TRUE : Pt_FALSE, Pt_BLOCKED );
		PtSetArg( &args[ 1 ], Pt_ARG_CURSOR_TYPE,
			(state) ? Ph_CURSOR_CLOCK : Ph_CURSOR_POINTER, 0 );
		PtSetResources( wgt, 2, args );
	}

}	// blockWindow


int
invit_setup( PtWidget_t *link_instance, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )
{

	/* eliminate 'unreferenced' warnings */
	link_instance = link_instance, apinfo = apinfo, cbinfo = cbinfo;

	/* Block the operation windows */
	blockWindow( ABW_chatrooms, TRUE );
	blockWindow( ABW_room, TRUE );

	/* Clear who to invit */
	setString( ABW_invit_who, "" );
	setString( ABW_invit_node, "" );

	/* Give focus to the 'Who' field */
	PtContainerGiveFocus( ABW_invit_who, cbinfo->event );

	/* Net stat */
	qnx_osinfo( 0, &odata );
	if ( nid_stat != NULL )
		free( nid_stat );
	nid_stat = calloc( odata.max_nodes+1, 1 );
	qnx_net_alive( nid_stat, odata.max_nodes+1 );
	
	return( Pt_CONTINUE );

}	// invit_setup


int
invit_destroyed( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )
{

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	/*--- Block the operation windows ---*/
	blockWindow( ABW_chatrooms, FALSE );
	blockWindow( ABW_room, FALSE );

	/*--- This dialog is no more active ---*/
	ABW_invit = NULL;
	
	/* Give focus to the input box */
	PtContainerGiveFocus( ABW_room_input, cbinfo->event );
	PtWindowFocus( ABW_room );
	PtWindowToFront( ABW_room );

	return( Pt_CONTINUE );

}	// invit_destroyed


int
invit_opening( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )
{

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	/* Hide net progression search */
	PtUnrealizeWidget( ABW_invit_nodes_progr );
	PtUnrealizeWidget( ABW_invit_diag );

	return( Pt_CONTINUE );

}	// invit_opening


static int do_invit( const char *who, const nid_t node, const pid_t pid_session_leader )
{
	char 			cmd[ PATH_MAX + 1 ];
	pid_t			pid;
	char			tmp1[ 20 ], tmp2[ 20 ], tmp3[ 20 ];
	pid_t			vid;

	if (_cmdname(cmd) == NULL) {
		setString( ABW_invit_diag, "Invitation failed [errno=%d]", errno );
		PtRealizeWidget( ABW_invit_diag );
		return( Pt_CONTINUE );
	}	// if

	/* Is the chatrooms app running on the target node ? */
	if ( (vid = qnx_name_locate( node, CLIENT_NAME, 0, NULL )) != -1 ) {
		setString( ABW_invit_diag, "User [%s] on //%d is already running chatrooms",
			who, node );
		PtRealizeWidget( ABW_invit_diag );
		qnx_vc_detach( vid );
#if 0
		PtContainerGiveFocus( room_texts[ curr_chair ], NULL );
#else
		PtContainerGiveFocus( ABW_room_input, NULL );
#endif
		return( Pt_CONTINUE );
	}	// if

	setString( ABW_room_infos, "[%s] invited on node %d", who, node );
	PtFlush();

	/* Create the process on the asked node */	
	qnx_spawn_options.node = node;
	qnx_spawn_options.flags |= 
		(_SPAWN_BGROUND | _SPAWN_NOZOMBIE ); // | _SPAWN_NEWPGRP | _SPAWN_TCSETPGRP);
	sprintf( tmp1, "%d", getnid() );
	sprintf( tmp2, "%d", curr_room );
	sprintf( tmp3, "%d", pid_session_leader );
	pid = spawnl( P_NOWAITO, cmd, cmd, "GlopGlop", whoami, tmp1, tmp2, who, tmp3, NULL );
	if ( pid == -1 ) {
		setString( ABW_invit_diag, "FAILURE [%s] invited on //%d (errno=%d)", 
			who, node, errno );
		PtRealizeWidget( ABW_invit_diag );
		return( Pt_CONTINUE );
	}	// if

#if 0
		PtContainerGiveFocus( room_texts[ curr_chair ], NULL );
#else
		PtContainerGiveFocus( ABW_room_input, NULL );
#endif

	PtDestroyWidget( ABW_invit );

	return( Pt_CONTINUE );

}	// do_invit


static int do_search_user( void )
{
	char	tmp[ 20 ];
	nid_t	node;
	pid_t	session_leader;

	getString( ABW_invit_who, who );
	getString( ABW_invit_node, tmp );
	node = (nid_t)atoi( tmp );

	/* Verify that the node is correct */
	if ( (node < 0) || (node > odata.max_nodes) ) {
		setString( ABW_invit_diag, "Node %d incorrect", node );
		PtRealizeWidget( ABW_invit_diag );
		return( Pt_CONTINUE );
	}	// if

	if ( search_user_on_node( node, &session_leader ) == 0 )
		return do_invit( who, node, session_leader );

	return( Pt_CONTINUE );

}	// do_search_user


int record_found_user( const char *user_name, nid_t node )
{
	char	value[ MAXLEN_USER ];
	char	tmp[ 20 ];

	if ( PxConfigOpen( ini_fname, PXCONFIG_WRITE) == Pt_FALSE )
		return -1;

	PxConfigSection( CFG_SECTION_USERS );
	if ( (PxConfigReadString( CFG_SECTION_USERS, user_name, "", value, MAXLEN_USER ) == Pt_TRUE) &&
		 (atoi( value ) == node) ) {
		PxConfigClose();
		return 0;
	}	// if

	sprintf( tmp, "%d", node );
	PxConfigWriteString( CFG_SECTION_USERS, user_name, PXCONFIG_FMT_STRING, tmp );

	PxConfigClose();
	return 0;

}	// record_found_user


static nid_t record_search_user( char *user_name, pid_t *pid_session_leader )
{
	char	value[ MAXLEN_USER+1 ];
	nid_t	node;

	if ( PxConfigOpen( ini_fname, PXCONFIG_READ) == Pt_FALSE )
		return -1;

	PxConfigSection( CFG_SECTION_USERS );
	if ( PxConfigReadString( CFG_SECTION_USERS, user_name, "", value, MAXLEN_USER ) == Pt_FALSE ) {
		PxConfigClose();
		return -1;
	}	// if

	PxConfigClose();

	node = (nid_t)atoi( value );
	if ( search_user_on_node( node, pid_session_leader ) == 0 )
		return node;
	
	return -1;

}	// record_search_user


static int search_user_on_net( nid_t *node, pid_t *pid_session_leader )
{
	nid_t			nid, _node;
	pid_t			vid;
	int				sid;
	struct _sidinfo	sdata;
	PtArg_t 		args[4];
	char			tmp[ 80 ];
	struct _osinfo	node_odata;

	/* First, try to use the information stored in the .cfg file, if available */
	if ( (_node = record_search_user( who, pid_session_leader )) != -1 ) {
		*node = _node;
		return 0;
	}	// if

	PtRealizeWidget( ABW_invit_nodes_progr );

	PtSetArg( &args[0], Pt_ARG_GAUGE_MINIMUM, 1, 0 );
	PtSetArg( &args[1], Pt_ARG_GAUGE_MAXIMUM, odata.max_nodes, 0 );
	PtSetArg( &args[2], Pt_ARG_GAUGE_VALUE, 1, 0 );
	sprintf( tmp, " / %d", odata.max_nodes );
	PtSetArg( &args[3], Pt_ARG_GAUGE_VALUE_SUFFIX, tmp, 0 );
	PtSetResources( ABW_invit_nodes_progr, 4, args );

	/* Search the asked user on each node of the net */
	blockWindow( ABW_invit, TRUE );
	for ( nid = 1; nid <= odata.max_nodes; nid++ ) {

		PtSetArg( &args[0], Pt_ARG_GAUGE_VALUE, nid, 0 );
		PtSetResources( ABW_invit_nodes_progr, 1, args );
		PtFlush();
	
		if ( !nid_stat[ nid ] )
			continue;
		if ( (vid = qnx_vc_attach( nid, PROC_PID, 0, 0 )) == -1 )
			continue;
		qnx_osinfo( vid, &node_odata );

		/* For each session of this node */
		for ( sid = 1; sid <= node_odata.num_sessions; sid++ ) {
			if ( (sid = qnx_sid_query( vid, sid, &sdata )) == -1 )
				break;
			if ( !strcmp( sdata.name, who ) ) {
				qnx_vc_detach( vid );
				*node = nid;
				*pid_session_leader = sdata.pid;
				record_found_user( sdata.name, nid );
				blockWindow( ABW_invit, FALSE );
				return 0;
			}	// if
		}	// for (sid)
		
		qnx_vc_detach( vid );

	}	// for (nid)
	blockWindow( ABW_invit, FALSE );

	/* Not found */
	setString( ABW_invit_diag, "User [%s] not found on the net", who );
	PtRealizeWidget( ABW_invit_diag );
	
	return -1;

}	// search_user_on_net


static int search_user_on_node( const nid_t node, pid_t *pid_session_leader )
{
	pid_t			vid;
	int				sid;
	struct _sidinfo	sdata;

	/* Node alive ? */
	if ( !nid_stat[ node ] ) {
		setString( ABW_invit_diag, "Node %d not alive", node );
		PtRealizeWidget( ABW_invit_diag );
		return -1;
	}	// if

	/* Attach a vc to this node */
	if ( (vid = qnx_vc_attach( node, PROC_PID, 0, 0 )) == -1 ) {
		setString( ABW_invit_diag, "Unable to attach vc on node %d", node );
		PtRealizeWidget( ABW_invit_diag );
		return -1;
	}	// if

	/* For each session of this node */
	for ( sid = 1; sid <= odata.num_sessions; sid++ ) {
		if ( (sid = qnx_sid_query( vid, sid, &sdata )) == -1 )
			break;
		if ( !strcmp( sdata.name, who ) ) {
			*pid_session_leader = sdata.pid;
			qnx_vc_detach( vid );
			record_found_user( sdata.name, node );
			return 0;
		}
	}	// for (sid)
		
	/* Not found */
	setString( ABW_invit_diag, "User [%s] not found on node %d", who, node );
	PtRealizeWidget( ABW_invit_diag );
	
	qnx_vc_detach( vid );
	return -1;

}	// search_user_on_node


static int try_to_locate_user_using_cfg( char *who, nid_t *node )
	{
	char	value[ MAXLEN_USER+1 ];

	if ( PxConfigOpen( ini_fname, PXCONFIG_READ) == Pt_FALSE )
		return -1;

	PxConfigSection( CFG_SECTION_USERS );
	if ( PxConfigReadString( CFG_SECTION_USERS, who, "", value, MAXLEN_USER ) == Pt_FALSE ) 
		{
		PxConfigClose();
		return -1;
		}	// if

	PxConfigClose();

	*node = (nid_t)atoi( value );
	return 0;

	}	// try_to_locate_user_using_cfg


int
invit_who_activate( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )
{
	PtTextCallback_t	*cbs = (PtTextCallback_t *)cbinfo->cbdata;
	char				tmp[ 20 ];
	nid_t				node;

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	if ( cbinfo->reason_subtype != Pt_EDIT_ACTIVATE ) {
		setString( ABW_invit_node, "??" );
		return( Pt_CONTINUE );
	}

	/* Remember the user name */
	strcpy( who, cbs->text );

	/*--- Is the node given ? ---*/
	getString( ABW_invit_node, tmp );
	if ( strcmp( tmp, "" ) )
		return do_search_user();
	
	/* Give focus to the node field */
	if ( try_to_locate_user_using_cfg( who, &node ) == 0 )
		setString( ABW_invit_node, "%d", node );
	PtContainerGiveFocus( ABW_invit_node, cbinfo->event );

	return( Pt_CONTINUE );

}	// invit_who_activate


int
invit_node_activate( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )
{
	PtTextCallback_t	*cbs = (PtTextCallback_t *)cbinfo->cbdata;
	nid_t				node;
	char				tmp[ 80 ];
	pid_t				session_leader;

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	getString( ABW_invit_who, tmp );
	if ( !strcmp( tmp, "" ) ) {
		PtContainerGiveFocus( ABW_invit_who, cbinfo->event );
		return( Pt_CONTINUE );
	}	// if

	/* No node means that we will search this user on the entire net */
	if ( (node = atoi( cbs->text )) == 0 ) {
		if ( search_user_on_net( &node, &session_leader ) == 0 )
			return do_invit( who, node, session_leader );
		return( Pt_CONTINUE );
	}	// if

	return do_search_user();

}	// invit_node_activate
