/* Y o u r   D e s c r i p t i o n                       */
/*                            AppBuilder Photon Code Lib */
/*                                         Version 1.13A */

/* Standard headers */
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <process.h>
#include <sys/sched.h>
#include <sys/qnx_glob.h>
#include <sys/name.h>
#include <unistd.h>
#include <sys/kernel.h>
#include <signal.h>
#include <sys/proxy.h>
#include <assert.h>
#include <sys/osinfo.h>
#include <sys/sidinfo.h>
#include <sys/vc.h>
#include <sys/types.h>
#include <pwd.h>

/* Toolkit headers */
#include <Ph.h>
#include <Pt.h>
#include <Ap.h>

/* Local headers */
#include "chatrooms.h"
#include "abimport.h"
#include "proto.h"

#include "chatrooms_vars.h"

extern	void show_users( void );
extern	void show_msg( void );

static	int				arm_cnt;


void setString (const PtWidget_t * wgt, const char *fmt,...)
{
	PtArg_t 	args[1];
	va_list 	ap;
	char 		buff[200];

	va_start (ap, fmt);
	vsprintf (buff, fmt, ap);
	va_end (ap);

	PtSetArg (&args[0], Pt_ARG_TEXT_STRING, buff, 0);
	PtSetResources (wgt, 1, args);
}	// setString


void getString (const PtWidget_t * wgt, char *string)
{
	PtArg_t 	args[1];
	char		*value;

	PtSetArg (&args[0], Pt_ARG_TEXT_STRING, &value, 0);
	PtGetResources (wgt, 1, args);
	strcpy (string, value);
}	// getString


int enableBtn (const PtWidget_t * wgt, const int state)
{
	PtArg_t args[3];

	PtSetArg (&args[0], Pt_ARG_FLAGS, (state) ? Pt_TRUE  : Pt_FALSE, Pt_SELECTABLE);
	PtSetArg (&args[1], Pt_ARG_FLAGS, (state) ? Pt_FALSE : Pt_TRUE,  Pt_GHOST);
	PtSetArg (&args[2], Pt_ARG_FLAGS, (state) ? Pt_FALSE : Pt_TRUE,  Pt_BLOCKED);
	PtSetResources (wgt, 3, args);
	return 0;

}	// enableBtn


int
room_setup( PtWidget_t *link_instance, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )
{
	USER			*p;
	int				chair;
	char			title[ 50 ];
	PtArg_t 		args[ 4 ];
	unsigned long	*flags;
	int				is_realized;

	/* eliminate 'unreferenced' warnings */
	link_instance = link_instance, apinfo = apinfo, cbinfo = cbinfo;

#define DEF( chair ) \
	room_chairs[ chair ] = ABW_room_chair##chair; \
	room_labels[ chair ] = ABW_room_who##chair; \
	room_texts[ chair ]  = ABW_room_text##chair
	DEF( 0 ); DEF( 1 ); DEF( 2 ); DEF( 3 ); DEF( 4 );

	/* Update the chairs */
	update_chairs( all_users[ curr_room ] );

	for ( p = all_users[ curr_room ], chair = 0; chair < MAXNB_CHAIRS; chair++ ) {
		active_text( chair, p, strcmp( p, "" ) );
		p++;
	}
#undef DEF

	/* Show all msg */
	show_msg();

	arm_cnt = 0;

	sprintf( title, "Chat room #%d", curr_room+1 );
	PtSetArg( &args[0], Pt_ARG_WINDOW_TITLE, title, 0 );
	PtSetResources( ABW_room, 1, args );

#if 0
	PtContainerGiveFocus( room_texts[ curr_chair ], NULL );
#else
	PtContainerGiveFocus( ABW_room_input, NULL );
#endif

	/* Disable access to the others rooms */
	PtSetArg( &args[0], Pt_ARG_FLAGS, Pt_FALSE, Pt_SELECTABLE );
	PtSetArg( &args[1], Pt_ARG_FLAGS, Pt_TRUE,  Pt_BLOCKED );
	PtSetArg( &args[2], Pt_ARG_FLAGS, Pt_FALSE, Pt_SET );
	PtSetArg( &args[3], Pt_ARG_FLAGS, Pt_FALSE, Pt_ETCH_HIGHLIGHT );
	PtSetResources( ABW_chatroom1_sel, 4, args );
	PtSetResources( ABW_chatroom2_sel, 4, args );
	PtSetResources( ABW_chatroom3_sel, 4, args );
	PtSetResources( ABW_chatroom4_sel, 4, args );

	enableBtn( ABW_room_btn_clear, FALSE );
	enableBtn( ABW_room_btn_send, FALSE );

	/*--- If the chatrooms window is realized, set the btn 'chatrooms' ---*/
	if ( chatrooms_was_hidden )
		PtUnrealizeWidget( ABW_chatrooms );
	PtSetArg( &args[ 0 ], Pt_ARG_FLAGS, &flags, 0 );
	PtGetResources( ABW_chatrooms, 1, args );
	is_realized = (((*flags & Pt_REALIZED) == Pt_REALIZED) ? TRUE : FALSE);
	PtSetArg( &args[ 0 ], Pt_ARG_FLAGS, (is_realized) ? Pt_TRUE : Pt_FALSE, Pt_SET );
	PtSetResources( ABW_room_btn_rooms, 1, args );

	return( Pt_CONTINUE );

}	// room_setup


int send_msg( void *data, pid_t pid, void *msg, ushort_t size )
{
	REQ_TYPE_MSG	req;
	ANSWER_TYPE_MSG	answer;
	int 			status;

	/* eliminate 'unreferenced' warnings */
	data= data, pid= pid, msg= msg, size= size;

	/* Message sent to the server */
	req.hdr.type = RQ_TYPE_MSG;
	req.userno   = userno;
	strcpy( req.msg, all_chairs_msg[ curr_chair ] );
	status = Send( server_vid, &req, &answer, 
		sizeof( REQ_TYPE_MSG ) - (sizeof( MSG ) - (strlen( req.msg ) + 1)),
		sizeof( ANSWER_TYPE_MSG ) );
	assert( status == 0 );
	if ( answer.ack != ACK_OK )
		return( Pt_CONTINUE );

	/* Reset counter */
	arm_cnt = 0;

	return( Pt_CONTINUE );

}	// send_msg


int
room_destroyed( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )
{
	REQ_EXIT_ROOM		req;
	ANSWER_EXIT_ROOM	answer;
	int					status;
	PtArg_t 			args[ 2 ];

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	/* Message sent to the server */
	req.hdr.type = RQ_EXIT_ROOM;
	req.userno   = userno;
	status = Send( server_vid, &req, &answer, sizeof( REQ_EXIT_ROOM ), sizeof( ANSWER_EXIT_ROOM ) );
	assert( (status == 0) );
	if ( answer.ack != ACK_OK )
		return( Pt_CONTINUE );
	memmove( all_users, answer.users, sizeof( all_users ) );

	/* Release the timer used for msg sent on timeout */
	timer_delete( timer_send_msg );
	qnx_proxy_detach( proxy_send_msg );
	PtAppRemoveInput( NULL, send_msg_id );

	ABW_room = NULL;

	show_users();

	/* Enable access to all rooms */
	PtSetArg( &args[0], Pt_ARG_FLAGS, Pt_TRUE, Pt_SELECTABLE );
	PtSetArg( &args[1], Pt_ARG_FLAGS, Pt_FALSE,  Pt_BLOCKED );
	PtSetResources( ABW_chatroom1_sel, 2, args );
	PtSetResources( ABW_chatroom2_sel, 2, args );
	PtSetResources( ABW_chatroom3_sel, 2, args );
	PtSetResources( ABW_chatroom4_sel, 2, args );

	/*--- Show the chatrooms window if it was hidden ---*/
	if ( chatrooms_was_hidden )
		PtRealizeWidget( ABW_chatrooms );

	/*--- Give focus to chatrooms box ---*/
	PtWindowFocus( ABW_chatrooms );
	PtWindowToFront( ABW_chatrooms );

	return( Pt_CONTINUE );

}	// room_destroyed


int
text_changed( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )
{
	struct itimerspec		timerSpec;
	PtMultiTextCallback_t	*mcbs = (PtMultiTextCallback_t *)cbinfo->cbdata;
	PtTextCallback_t		*cbs  = &mcbs->tc;

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	/* Arm the timer to enable send msg on timeout */
	timerSpec.it_value.tv_sec = 0;					// 0 sec. before initial firing
	if ( ++arm_cnt == 10 )
		timerSpec.it_value.tv_nsec = 100000;
	else
		timerSpec.it_value.tv_nsec = 250000000L;	// 250 msec
	timerSpec.it_interval.tv_sec  = 0;				// repetitive timer afterwards
	timerSpec.it_interval.tv_nsec = 0;
	timer_settime( timer_send_msg, 0, &timerSpec, NULL );

	/* Update the text in local buffer */
	strcpy( all_chairs_msg[ curr_chair ], cbs->text );

	return( Pt_CONTINUE );

}	// text_changed


int
room_input_activate( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )
{
	PtMultiTextCallback_t	*mcbs = (PtMultiTextCallback_t *)cbinfo->cbdata;
	PtTextCallback_t		*cbs  = &mcbs->tc;
	REQ_TYPE_MSG			req;
	ANSWER_TYPE_MSG			answer;
	int 					status;
	struct itimerspec		timerSpec;

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	/* Update the text in local buffer */
	strcpy( all_chairs_msg[ curr_chair ], cbs->text );
	setString( room_texts[ curr_chair ], all_chairs_msg[ curr_chair ] );
	setString( ABW_room_input, "" );

	/* Message sent to the server */
	req.hdr.type = RQ_TYPE_MSG;
	req.userno   = userno;
	strcpy( req.msg, all_chairs_msg[ curr_chair ] );
	status = Send( server_vid, &req, &answer, 
		sizeof( REQ_TYPE_MSG ) - (sizeof( MSG ) - (strlen( req.msg ) + 1)),
		sizeof( ANSWER_TYPE_MSG ) );
	assert( status == 0 );
	if ( answer.ack != ACK_OK )
		return( Pt_CONTINUE );

	/* Reset timer */
	timerSpec.it_value.tv_sec     = 
	timerSpec.it_value.tv_nsec    = 
	timerSpec.it_interval.tv_sec  = 
	timerSpec.it_interval.tv_nsec = 0;
	timer_settime( timer_send_msg, 0, &timerSpec, NULL );
	arm_cnt = 0;

	/* Reset the string */
	enableBtn( ABW_room_btn_clear, FALSE );
	enableBtn( ABW_room_btn_send, FALSE );

	/* Give focus to the input box */
	PtContainerGiveFocus( ABW_room_input, cbinfo->event );

	return( Pt_CONTINUE );

}	// room_input_activate


int
room_input_changed( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )
{

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	enableBtn( ABW_room_btn_clear, TRUE );
	enableBtn( ABW_room_btn_send, TRUE );

	return( Pt_CONTINUE );

}	// room_input_changed


int
room_btn_clear( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )
{

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	enableBtn( ABW_room_btn_clear, FALSE );
	enableBtn( ABW_room_btn_send, FALSE );
	setString( ABW_room_input, "" );

	/* Give focus to the input box */
	PtContainerGiveFocus( ABW_room_input, cbinfo->event );

	return( Pt_CONTINUE );

}	// room_btn_clear


int
room_btn_send( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )
{
	REQ_TYPE_MSG			req;
	ANSWER_TYPE_MSG			answer;
	int 					status;
	struct itimerspec		timerSpec;

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	/* Update the text in local buffer */
	getString( ABW_room_input, all_chairs_msg[ curr_chair ] );
	setString( room_texts[ curr_chair ], all_chairs_msg[ curr_chair ] );
	setString( ABW_room_input, "" );

	/* Message sent to the server */
	req.hdr.type = RQ_TYPE_MSG;
	req.userno   = userno;
	strcpy( req.msg, all_chairs_msg[ curr_chair ] );
	status = Send( server_vid, &req, &answer, 
		sizeof( REQ_TYPE_MSG ) - (sizeof( MSG ) - (strlen( req.msg ) + 1)),
		sizeof( ANSWER_TYPE_MSG ) );
	assert( status == 0 );
	if ( answer.ack != ACK_OK )
		return( Pt_CONTINUE );

	/* Reset timer */
	timerSpec.it_value.tv_sec     = 
	timerSpec.it_value.tv_nsec    = 
	timerSpec.it_interval.tv_sec  = 
	timerSpec.it_interval.tv_nsec = 0;
	timer_settime( timer_send_msg, 0, &timerSpec, NULL );
	arm_cnt = 0;

	/* Reset the string */
	enableBtn( ABW_room_btn_clear, FALSE );
	enableBtn( ABW_room_btn_send, FALSE );

	/* Give focus to the input box */
	PtContainerGiveFocus( ABW_room_input, cbinfo->event );

	return( Pt_CONTINUE );

}	// room_btn_send


int
room_btn_rooms( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )
	{
	PtArg_t			args[ 1 ];
	unsigned long	*flags;
	int				is_set;

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	/*--- Is the button set or not ? ---*/
	PtSetArg( &args[ 0 ], Pt_ARG_FLAGS, &flags, 0 );
	PtGetResources( widget, 1, args );
	is_set = (((*flags & Pt_SET) == Pt_SET) ? TRUE : FALSE);

	/*--- Hide or show the main chatrooms window */
	if ( !is_set )
		PtUnrealizeWidget( ABW_chatrooms );
	else
		PtRealizeWidget( ABW_chatrooms );
	PtWindowToFront( ABW_room );
	PtSetArg( &args[ 0 ], Pt_ARG_FLAGS, &flags, 0 );
	PtGetResources( ABW_chatrooms, 1, args );
	chatrooms_was_hidden = (((*flags & Pt_REALIZED) == Pt_REALIZED) ? FALSE : TRUE);

	return( Pt_CONTINUE );

	}	// room_btn_rooms


int
room_realized( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )
	{

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	return install_balloons( widget );

	}	// room_realized



int
room_unrealized( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )
	{

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	return desinstall_balloons( widget );

	}	// room_unrealized

